/******************************************************************************
Description:

  Example CIF program to test and demonstrate the C++ implementation

CVS Information:

  $Source$
  $Date$
  $Revision$
  $State$
  $Author$
******************************************************************************/

/******************************************************************************
			     Include header files
******************************************************************************/
#include "CIF.H"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>


/******************************************************************************
			     Function definitions
******************************************************************************/

/*
 * usage()
 */
void
usage(
    char *				exec_name)
{
    fprintf(stderr, "\nusage: %s [-c <nmsgs>] [-f] [-l] [-t <timeout>] "
	    "<-s|-r> <URL>\n\n", exec_name);
    fprintf(stderr, "    -c\tmessage count\n");
    fprintf(stderr, "    -f\trepeat forever\n");
    fprintf(stderr, "    -l\tlistener\n");
    fprintf(stderr, "    -r\treciever\n");
    fprintf(stderr, "    -s\tsender\n");
    fprintf(stderr, "    -t\ttimeout value\n\n");

    exit(1);
}


/*
 * main()
 */
int    
main(
    int					argc,
    char *				argv[])
{
    const int				MAX_MSG_SZ = 512;

    int					argn;
    int					msg_cnt = 1;
    CIF_Bool				forever = false;
    CIF_Bool				listener = false;
    CIF_Bool				sender = false;
    CIF_Bool				receiver = false;
    int					timeout = -1;    
    char *				url;

    /*
     * Process arguments
     */
    argn = 1;
    while(argn < argc - 1)
    {
	if (strcmp(argv[argn], "-c") == 0)
	{
	    msg_cnt = atoi(argv[++argn]);
	}
	else if (strcmp(argv[argn], "-f") == 0)
	{
	    forever = true;
	}
	else if (strcmp(argv[argn], "-l") == 0)
	{
	    listener = true;
	}
	else if (strcmp(argv[argn], "-s") == 0)
	{
	    sender = true;
	}
	else if (strcmp(argv[argn], "-r") == 0)
	{
	    receiver = true;
	}
	else if (strcmp(argv[argn], "-t") == 0)
	{
	    timeout = atoi(argv[++argn]);
	}
	else
	{
	    fprintf(stderr, "unknown argument: %s\n\n", argv[argn]);
	    usage(argv[0]);
	}

	argn++;
    }

    /*
     * Verify that all of the necessary arguments were specified
     */
    if (argc > 1)
    {
	url = new char [strlen(argv[argn])+1];
	strcpy(url, argv[argn]);
    }
    else
    {
	usage(argv[0]);
    }

    if (!sender && !receiver)
    {
	usage(argv[0]);
    }
    

    /*
     * Activate CIF
     *
     * CIF::activate() must be executed before any other CIF method is called.
     */
    CIF::activate();

    /*
     * Establish a listener object and the wait for an incoming connection
     */
    CIF_Listener *			cif_listener;
    CIF_Bool				rc;

    if (listener)
    {
	printf("Establishing listener object for URL=%s\n", url);
	cif_listener = CIF_Factory::get_Listener(url);
	assert(CIF::status == CIF::STATUS_SUCCESS);
	assert(cif_listener != CIF_NULL);

	printf("\nListening at %s\n\n", cif_listener->get_URL());
    }
    

    do
    {
	CIF_Connection *		cif_conn;

	if (listener)
	{
	    /*
	     * Wait for a remote process contact the listener and form a
	     * new connection.
	     */
	    printf("Waiting for connection\n");
	    if (timeout >= 0)
	    {
		cif_conn = cif_listener->get_Connection(timeout);
	    }
	    else
	    {
		cif_conn = cif_listener->get_Connection();
	    }
	    
	    assert(CIF::status == CIF::STATUS_SUCCESS);
	    assert(cif_conn != CIF_NULL);
	}
	else
	{
	    /*
	     * Attempt to establish a new connection
	     */
	    printf("Establishing connection object for URL=%s\n", url);
	    if (timeout >= 0)
	    {
		cif_conn = CIF_Factory::get_Connection(url, timeout);
	    }
	    else
	    {
		cif_conn = CIF_Factory::get_Connection(url);
	    }
	    assert(CIF::status == CIF::STATUS_SUCCESS);
	    assert(cif_conn != CIF_NULL);
	}
	printf("Connection established.\n");
	
	int				cnt;

	if (sender)
	{
	    /*
	     * Send msg_cnt number of messages
	     */
	    printf("Sending %d messages\n", msg_cnt);
	    for (cnt = 0; cnt < msg_cnt; cnt++)
	    {
		CIF_Byte		buf[MAX_MSG_SZ];
	    
		sprintf((char *) buf, "data for message %d", cnt);
		rc = cif_conn->send(buf, strlen((char *) buf) + 1, cnt);
		assert(rc == true);
		assert(CIF::status == CIF::STATUS_SUCCESS);
	    }
	}
	else
	{
	    /*
	     * Wait for the first message to be received.  This is not
	     * necessary since receive will block until a message is receive;
	     * however, it demostrates the use of the poll() method.
	     */
	    while (cif_conn->poll() == false)
	    {
		assert(CIF::status == CIF::STATUS_SUCCESS);
	    }

	    /*
	     * Revceive msg_cnt number of messages
	     */
	    for (cnt = 0; cnt < msg_cnt; cnt++)
	    {
		CIF_Message		message;
	    
		rc = cif_conn->receive(message);
		assert(rc == true);
		assert(CIF::status == CIF::STATUS_SUCCESS);

		printf("message received, size=%d, tag=%d, data=%s\n",
		       message.get_Info().size,
		       message.get_Info().tag,
		       message.get_Data());
	    
	    }
	}

	/*
	 * Close the connection and delete the connection object
	 */
	printf("Communication complete.  Closing connection.\n");
	rc = cif_conn->close();
	assert(CIF::status == CIF::STATUS_SUCCESS);
	assert(rc == true);

	delete cif_conn;
    }
    while (forever);

    /*
     * Close the listener and delete the listener object
     */
    if (listener)
    {
	printf("Closing listener.\n");
	rc = cif_listener->close();
	assert(CIF::status == CIF::STATUS_SUCCESS);
	assert(rc == true);

	delete cif_listener;
    }
    
    /*
     * Detactivate CIF
     *
     * CIF::deactivate() must be called when the application has finished using
     * CIF.  Normally, this is done right before the application exits.
     */
    CIF::deactivate();

    exit(0);
}


