/*
 * harness_endpoint.c
 */

#include <nexus.h>
#include "harness_endpoint.h"
#include "ep_list.h"
#include "sp_group.h"
#include "string.h"
#include "startup.h"

static nexus_handler_t handler_table[] =
{
    {NEXUS_HANDLER_TYPE_NON_THREADED,	attach_startpoints_handler},
    {NEXUS_HANDLER_TYPE_NON_THREADED,	startup_handler},
    {NEXUS_HANDLER_TYPE_NON_THREADED,	shutdown_handler},
    {NEXUS_HANDLER_TYPE_NON_THREADED,	request_sp_copies_handler},
    {NEXUS_HANDLER_TYPE_NON_THREADED,	collect_sp_copies_handler},
    {NEXUS_HANDLER_TYPE_NON_THREADED,	scatter_sp_copies_handler},
    {NEXUS_HANDLER_TYPE_NON_THREADED,	ping_handler},
    {NEXUS_HANDLER_TYPE_NON_THREADED,	barrier_handler},
    {NEXUS_HANDLER_TYPE_NON_THREADED,	option_string_handler},
    {NEXUS_HANDLER_TYPE_NON_THREADED,	set_debug_level_handler}
};
#define handler_table_size	(sizeof(handler_table) / \
				 sizeof(nexus_handler_t))
static nexus_endpoint_t harness_endpoint;

static struct
{
    nexus_mutex_t	mutex;
    nexus_cond_t	cond;
    int			done;
} monitor;

static int	next_ping_id = 1;

static int
unique_ping_id()
{
    int	x;

    x = next_ping_id++;
    x += 100000 * (sp_group_rank(HARNESS_GROUP_ID) + 1);
    return (x);
}

static void
cleanup_func(nexus_endpoint_t *ep)
{
    nexus_mutex_destroy(&(monitor.mutex));
    nexus_cond_destroy(&(monitor.cond));
}

int
harness_endpoint_init()
{
    nexus_endpointattr_t	ep_attr;

    if (nexus_endpointattr_init(&ep_attr) != NEXUS_SUCCESS) {
	nexus_printf("harness_endpoint_init(): attr_init failed\n");
	return (1);
    }
    if (nexus_endpointattr_set_handler_table(&ep_attr,
					     handler_table,
					     handler_table_size)
	!= NEXUS_SUCCESS) {
	nexus_printf("harness_endpoint_init(): set_table failed\n");
	return (1);
    }
    if (nexus_endpoint_init(&harness_endpoint, &ep_attr)
	!= NEXUS_SUCCESS) {
	nexus_printf("harness_endpoint_init(): ep_init failed\n");
	return (1);
    }
    nexus_endpointattr_destroy(&ep_attr);
    if (ep_list_new_ep(&harness_endpoint,
		       NEXUS_FALSE,
		       cleanup_func) != 0) {
	nexus_printf("harness_endpoint_init(): new_ep failed\n");
	return (1);
    }
    /* borrowing the test debug for just this one */
    TraceTst(nexus_printf("harness_endpoint_init(): success\n"));
    return (0);
}

void
harness_send_ping(int dest, int trips,
		  nexus_bool_t is_non_threaded_handler)
{
    nexus_buffer_t	buf;
    int			src, pid;

    if (trips > 0) {
	trips = trips * 2 - 1;	/* round trips */
	monitor.done = 1;
    } else
	trips = -1;	/* one way */
    src = sp_group_rank(HARNESS_GROUP_ID);
    pid = unique_ping_id();
    nexus_buffer_init(&buf, 4 * nexus_sizeof_int(1), 0);
    nexus_put_int(&buf, &src, 1);
    nexus_put_int(&buf, &dest, 1);
    nexus_put_int(&buf, &pid, 1);
    nexus_put_int(&buf, &trips, 1);
    nexus_printf("Initiating ping: src=%i,pid=%i,dst=%i\n",
		 src, pid, dest);
    nexus_send_rsr(&buf,
		   sp_group_sp(HARNESS_GROUP_ID, dest),
		   PING_HANDLER_ID,
		   NEXUS_TRUE,
		   is_non_threaded_handler);
    if (trips > 0) {
	nexus_mutex_lock(&(monitor.mutex));
	while (monitor.done != 0)
	    nexus_cond_wait(&(monitor.cond),
			    &(monitor.mutex));
	nexus_mutex_unlock(&(monitor.mutex));
	nexus_printf("ping %i complete\n", pid);
    }
}

void
harness_ping_test(int n_trips, int n_rounds)
{
    int		i, j;

    for (i = 0; i < n_rounds; i++)
	for (j = 0; j < sp_group_size(HARNESS_GROUP_ID); j++)
	    harness_send_ping(j, n_trips, NEXUS_FALSE);
}

/*
 *
 * handlers
 *
 */

void
ping_handler(nexus_endpoint_t *endpoint,
		  nexus_buffer_t *buffer,
		  nexus_bool_t is_non_threaded_handler)
{
    nexus_buffer_t	reply_buf;
    int			src, dst, pid, ttl, x;

    nexus_get_int(buffer, &src, 1);
    nexus_get_int(buffer, &dst, 1);
    nexus_get_int(buffer, &pid, 1);
    nexus_get_int(buffer, &ttl, 1);
    nexus_printf("ping_send_handler(): dst=%i,pid=%i,src=%i,ttl=%i\n",
		 dst, pid, src, ttl);
    x = sp_group_rank(HARNESS_GROUP_ID);
    if (dst != x)
	nexus_printf("ping_send_handler(): I am %i, not %i\n",
		     x, dst);
    
    if (ttl == -1) {
	/* one way, let it die */
	return;
    } else if (ttl == 0) {
	/* end of trip, signal */
	nexus_mutex_lock(&(monitor.mutex));
	monitor.done = 0;
	nexus_cond_signal(&(monitor.cond));
	nexus_mutex_unlock(&(monitor.mutex));
    } else {
	/* send reply */
	ttl--;
	nexus_buffer_init(&reply_buf, 4 * nexus_sizeof_int(1), 0);
	nexus_put_int(&reply_buf, &dst, 1); /* new src */
	nexus_put_int(&reply_buf, &src, 1); /* new dst */
	nexus_put_int(&reply_buf, &pid, 1);
	nexus_put_int(&reply_buf, &ttl, 1);
	nexus_send_rsr(&reply_buf,
		       sp_group_sp(HARNESS_GROUP_ID, src),
		       PING_HANDLER_ID,
		       NEXUS_TRUE,
		       is_non_threaded_handler);
    }
}
